// A metaphor.
use crate::{math, maze};

static mut MOVE_SPEED: u32 = 0;
static mut NOTICE_SPEED: u32 = 0;

static mut FROM: (u8, u8) = (0, 0);
static mut TO: (u8, u8) = (0, 0);
static mut GOAL: (u8, u8) = (0, 0);
static mut MOVE_PROGRESS: u32 = 0;
static mut NOTICE_PROGRESS: u32 = 0;
static mut CHASING: bool = false;

const PALETTE_WANDER: [u32; 4] = [0x96acd6, 0x6f9be2, 0x3b59d0, 0x0c1429];
const PALETTE_SEARCH: [u32; 4] = [0x6f9be2, 0x3b59d0, 0x0c1449, 0x191020];
const PALETTE_CHASE: [u32; 4] = [0xf2e696, 0xad2020, 0x560909, 0x190000];

pub unsafe fn initialize(pos: (u8, u8), difficulty: u8) {
    FROM = pos;
    TO = pos;
    GOAL = pos;

    MOVE_SPEED = 40 - 2 * difficulty as u32;
    NOTICE_SPEED = 50 - 3 * difficulty as u32;

    MOVE_PROGRESS = MOVE_SPEED;
    NOTICE_PROGRESS = NOTICE_SPEED;
    CHASING = false;
}

unsafe fn get_position() -> (f32, f32) {
    let t = MOVE_PROGRESS as f32 / MOVE_SPEED as f32;
    let x = FROM.0 as f32 * (1.0 - t) + TO.0 as f32 * t;
    let y = FROM.1 as f32 * (1.0 - t) + TO.1 as f32 * t;
    (x, y)
}

pub unsafe fn get_palette() -> [u32; 4] {
    if CHASING {
        PALETTE_CHASE
    } else if NOTICE_PROGRESS == NOTICE_SPEED {
        PALETTE_WANDER
    } else {
        PALETTE_SEARCH
    }
}

pub unsafe fn update(m: &maze::Board, player: Option<(f32, f32)>) -> (f32, f32) {
    MOVE_PROGRESS += 1;
    if MOVE_PROGRESS >= MOVE_SPEED {
        MOVE_PROGRESS = 0;
        FROM = TO;
        if TO == GOAL {
            GOAL = m.random_point();
            CHASING = false;
        }
        TO = m.nav(FROM, GOAL).travel(FROM);
    }
    let pos = get_position();
    if let Some(target) = player {
        if m.can_see(pos, target) {
            NOTICE_PROGRESS = NOTICE_PROGRESS.saturating_sub(3);
            if NOTICE_PROGRESS == 0 {
                GOAL = (math::round(target.0) as u8, math::round(target.1) as u8);
                CHASING = true;
            }
        } else {
            NOTICE_PROGRESS = NOTICE_SPEED.min(NOTICE_PROGRESS + 1);
        }
    } else {
        NOTICE_PROGRESS = NOTICE_SPEED.min(NOTICE_PROGRESS + 2);
    }
    pos
}

pub unsafe fn is_player_eaten(player: (f32, f32)) -> bool {
    if !CHASING {
        return false;
    }
    let pos = get_position();
    let x = pos.0 - player.0;
    let y = pos.1 - player.1;
    math::sqrt(x * x + y * y) < 8.0
}
