// Scrumptious. Surprising.
pub use core::f32::consts::PI;
pub const HALF_PI: f32 = 0.5 * PI;
pub const TWO_PI: f32 = 2.0 * PI;

pub fn sin(x: f32) -> f32 {
    let r = wrap_angle(x);
    let _sin = |x| 16.0 * x * (PI - x) / (5.0 * PI * PI - 4.0 * x * (PI - x));
    if r > PI {
        -_sin(r - PI)
    } else {
        _sin(r)
    }
}

pub fn cos(x: f32) -> f32 {
    sin(x + HALF_PI)
}

pub fn _tan(x: f32) -> f32 {
    sin(x) / cos(x)
}

pub fn atan2(y: f32, x: f32) -> f32 {
    let ax = abs(x);
    let ay = abs(y);
    let a = ax.min(ay) / ax.max(ay);
    let s = a * a;
    let mut r = ((-0.046_496_473 * s + 0.15931422) * s - 0.327_622_77) * s * a + a;
    if ay > ax {
        r = HALF_PI - r;
    }
    if x < 0.0 {
        r = PI - r;
    }
    if y < 0.0 {
        r = -r;
    }
    r
}

pub fn sqrt(v: f32) -> f32 {
    let mut guess = v;
    for _ in 0..100 {
        guess = sqrt_iterate(guess, v);
    }
    guess
}

fn sqrt_iterate(x: f32, s: f32) -> f32 {
    x + s / x
}

pub fn abs(v: f32) -> f32 {
    if v < 0.0 {
        -v
    } else {
        v
    }
}

pub fn round(v: f32) -> i8 {
    (v + 0.5) as i8
}

pub fn euclid_mod(v: f32, divisor: f32) -> f32 {
    let r = v % (divisor);
    if r < 0.0 {
        r + divisor
    } else {
        r
    }
}

pub fn wrap_angle(a: f32) -> f32 {
    euclid_mod(a, TWO_PI)
}
