// TODO
use crate::{math, maze, wasm4};
use math::*;
use maze::Board;
use wasm4::*;

type FloatPos = (f32, f32);

const FOV: f32 = PI * 0.4;
const SCALE: f32 = 2.0;
const WIDTH: u32 = SCREEN_SIZE;
const HEIGHT: u32 = SCREEN_SIZE;
const X_OFFSET: u32 = 0;
const Y_OFFSET: u32 = 0;

static mut DEPTH_BUFFER: [f32; WIDTH as usize] = [0.0; WIDTH as usize];

fn change_line_color(index: u32) {
    let val = index as u16 + 1;
    unsafe {
        *DRAW_COLORS = (val << 4) + val;
    }
}

pub fn draw_maze(b: &Board, pos: FloatPos, angle: f32) {
    //trace("what");
    change_line_color(3);
    rect(X_OFFSET as i32, Y_OFFSET as i32, WIDTH, HEIGHT);
    for x in 0..WIDTH {
        let ray_angle: f32 = wrap_angle(angle - FOV / 2.0) + (x as f32 / WIDTH as f32) * FOV;

        let eye_x = cos(ray_angle);
        let eye_y = sin(ray_angle);

        let mut distance: f32 = 0.0;
        for _ in 0..601 {
            distance += 0.01;
            let test_x = round(pos.0 + eye_x * distance) as u8;
            let test_y = round(pos.1 + eye_y * distance) as u8;
            if b.get_unbounded(test_x, test_y) {
                break;
            }
        }
        unsafe {
            DEPTH_BUFFER[x as usize] = distance;
        }
        let line_color = if distance < 1.5 {
            0
        } else if distance < 3.0 {
            1
        } else if distance < 6.0 {
            2
        } else {
            3
        };
        let wall_height = (HEIGHT as f32 / (distance * SCALE * 2.0)) as u32;
        let clipped_height = wall_height.min(HEIGHT / 2);
        let y_center = HEIGHT / 2 + Y_OFFSET;

        change_line_color(line_color);
        let ux = (x + X_OFFSET) as i32;
        vline(ux, (y_center - clipped_height) as i32, 2 * clipped_height);
    }
}

pub fn world_to_local(object: FloatPos, camera: FloatPos, camera_angle: f32) -> FloatPos {
    let o = (object.0 - camera.0, object.1 - camera.1);
    let dist = o.0 * o.0 + o.1 * o.1;
    let angle = math::atan2(o.1, o.0) - camera_angle;
    (dist * math::cos(angle), dist * math::sin(angle))
}

pub fn draw_sprite(pixels: &[u8], w: u32, h: u32, pos: FloatPos) {
    let spw = w as f32;
    let dist = sqrt(pos.0 * pos.0 + pos.1 * pos.1);
    let sprite_width = (spw * SCREEN_SIZE as f32 / (dist * SCALE * 2.0)) as i32;
    let angle = atan2(pos.1, pos.0);
    let lerp_angle = WIDTH as f32 * angle / FOV;
    let screen_pos = (WIDTH as f32 / 2.0 + lerp_angle) as i32;
    for raw_x in 0..sprite_width {
        let x = raw_x + screen_pos - sprite_width / 2;
        if x < 0 || x >= WIDTH as i32 {
            continue;
        }
        let depth = unsafe { &mut DEPTH_BUFFER[x as usize] };
        // No, I don't know.
        if *depth * 32.0 < dist {
            continue;
        }
        *depth = dist;
        let sample_x = raw_x as f32 / sprite_width as f32;
        for raw_y in 0..sprite_width {
            let sample_y = raw_y as f32 / sprite_width as f32;
            if sample_sprite(pixels, w, h, sample_x, sample_y) {
                let y = raw_y - sprite_width / 2 + 80;
                pixel(x, y);
            }
        }
    }
}

pub fn sample_sprite(pixels: &[u8], w: u32, h: u32, x: f32, y: f32) -> bool {
    let u = x * w as f32;
    let v = y * h as f32;
    let index = v as usize * w as usize + u as usize;
    let mask = 128 >> (index % 8);
    pixels[index / 8] & mask != 0
}
